""" This is the CARTA launcher. """
import logging
import os
import re
import signal
import subprocess
import sys
from pathlib import Path
from types import FrameType
from typing import Optional

# pylint: disable=E0401, R0913, R1721, W0603, W1203

from carta_envoy.connect import ArchiveConnect, NotificationConnect, RedisConnect
from carta_envoy.utilities import CARTA_URL_REPLACE_TEXT, CARTA_HTML_FILENAME, CARTA_TEMPLATE

"""
Setup and Launch a CARTA session
"""

CARTA_PROCESS: Optional[subprocess.Popen] = None


class CartaLauncher:
    """Runs CARTA"""

    def __init__(self, settings: dict):
        self.logger = logging.getLogger("carta_envoy")
        self.settings = settings
        self.signal = self.signal_handler
        self.redis_connect = RedisConnect(settings)
        self.archive_connect = ArchiveConnect(settings)
        self.notification = NotificationConnect(settings)

    def signal_handler(self, signum: int, frame: FrameType):
        """
        Handle a signal received by tearing down the program and exiting
        NOTE: Parameters signum and frame are convention for signal handlers,
        even if they don't end up used

        :param signum: Signal number
        :param frame: Stack frame
        """
        if signum == signal.SIGINT:
            self.logger.warning("WARNING: SIGINT received. Exiting.")
        elif signum == signal.SIGALRM:
            self.logger.warning("WARNING: Timeout elapsed. Exiting.")
        self.teardown()
        sys.exit()

    def prepare_for_carta(self):
        """
        Sets Redis values necessary for CARTA.

        :return:
        """
        ports_and_url = self.redis_connect.prepare_redis()
        signal.signal(signal.SIGINT, self.signal)
        return ports_and_url

    def launch(self):
        """
        Launches CARTA.

        :return:
        """
        self.logger.info("RUNNING CARTA!")
        prepare = self.prepare_for_carta()

        self.run_carta(
            self.settings["carta_path"],
            int(self.settings["timeout"]),
            Path(self.settings["data_location"]),
            prepare["front_end_port"],
            prepare["back_end_port"],
            prepare["carta_url"],
        )

    def run_carta(
        self,
        path_to_carta: str,
        timeout_minutes: int,
        file_browser_path: Path,
        front_end_port: int,
        back_end_port: int,
        carta_url: str,
    ):
        """
        Runs CARTA executable. Will close after a certain amount of time.

        :param path_to_carta: Path to CARTA executable
        :param timeout_minutes: Number of minutes to execute CARTA for before stopping it
        :param file_browser_path: Path to use for CARTA file browser
        :param front_end_port: Port to run the CARTA front end with
        :param back_end_port: Port to run the CARTA back end with
        :param carta_url: URL used to access CARTA from a web browser
        """
        global CARTA_PROCESS

        self.logger.info(f"file_browser_path: {file_browser_path}")
        if not file_browser_path.exists():
            self.logger.error(
                f"ERROR: Invalid file browser path. {file_browser_path}: Directory does not exist."
            )
            return
        elif not file_browser_path.is_dir():
            self.logger.error(
                f"ERROR: Invalid file browser path. {file_browser_path}: Path is not a directory."
            )
            return

        if self.settings["useCarta"]:
            self.logger.info(f"Running CARTA with a timeout of {timeout_minutes} minute(s)...")
            if re.search(r"CARTA.AppImage", path_to_carta):
                # Command uses AppImage, include --remote option
                carta_command = [
                    path_to_carta,
                    "--remote",
                    f"--port={str(back_end_port)}",
                    f"--fport={str(front_end_port)}",
                    f"--folder={file_browser_path}",
                    f"--root={file_browser_path}",
                ]
            else:
                carta_command = [
                    path_to_carta,
                    f"--port={str(back_end_port)}",
                    f"--fport={str(front_end_port)}",
                    f"--folder={file_browser_path}",
                    f"--root={file_browser_path}",
                ]
            try:
                CARTA_PROCESS = subprocess.Popen(
                    carta_command,
                    preexec_fn=os.setpgrp(),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            except (OSError, FileNotFoundError, subprocess.SubprocessError) as err:
                self.teardown()
                sys.exit(f"ERROR: Failed to launch CARTA: {err}")
            else:
                self.create_frame_html(carta_url=carta_url, html_dir=file_browser_path)

                # CARTA is running and accessible, so send CARTA URL to AAT system or notify user
                self.notify_ready(carta_url=carta_url, html_dir=file_browser_path)

                # Activate timeout handler
                signal.signal(signal.SIGALRM, self.signal)
                signal.alarm(60 * timeout_minutes)
                self.logger.info(f"To access CARTA, enter this URL into a web browser: {carta_url}")
                self.logger.info("Press Ctrl+C to quit.")

                # Loop until process ends or timeout elapses, whichever happens first
                while CARTA_PROCESS.returncode is None:
                    CARTA_PROCESS.poll()
        else:
            self.logger.info("Running locally...")

    def notify_ready(self, carta_url: str, html_dir: Path):
        """
        Sends URL notification to user and request handler

        :param carta_url: URL to CARTA session
        :param html_dir: location of CARTA HTML wrapper

        :return:
        """


        self.logger.info("SENDING URL NOTIFICATION TO: ")
        if self.settings["send_ready"] == "true":
            self.logger.info(f"User Email")
            self.notification.send_session_ready(carta_url)
        elif self.settings["send_ready"] == "false":
            self.logger.info(f"AAT Request Handler, with {carta_url}")
            self.archive_connect.send_carta_url_to_rh(carta_url)


    def teardown(self):
        """
        Deletes instance's Redis keys from the Redis server
        """
        self.notification.send_session_expired()

        self.logger.info("Cleaning up...")
        self.redis_connect.remove_redis_values()

        if CARTA_PROCESS and CARTA_PROCESS.returncode is None:
            # If CARTA is running, kill it
            try:
                # End the process (simulate Ctrl+Z)
                CARTA_PROCESS.send_signal(signal.SIGSTOP)
                # Kill the process
                os.killpg(os.getpgid(CARTA_PROCESS.pid), signal.SIGTERM)
                CARTA_PROCESS.wait()
            except OSError:
                # Process isn't running; just pass
                pass
            else:
                self.logger.info("Killed running CARTA process.")
        else:
            self.logger.warning("WARNING: CARTA not running.")

    @staticmethod
    def create_frame_html(carta_url: str, html_dir: Path) -> Path:
        """
        Generate the HTML page containing the CARTA URL in a frame.

        :param carta_url: URL of CARTA session
        :param html_dir: where HTML will be written
        :return: HTML file we just created
        """

        new_content = CARTA_TEMPLATE.replace(CARTA_URL_REPLACE_TEXT, carta_url)
        html_file = html_dir / CARTA_HTML_FILENAME
        html_file.write_text(new_content)

        return html_file
